/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "pim/AppointmentBuilder.h"
#include "base/Log.h"
#include "pim/difftime.h"
#include "pim/UtilsRecurrence.h"
#include "pim/MAPIUtils.h"
#include "base/util/utils.h"
#include "syncml/core/PropParam.h"


/**
 * Enumeration of appointment fields used by WM plugin.
 * @note Don't change the order of these properties!
 *       This list MUST be synchronized with PIMPR_props_app array of CEPROPID fields.
 */
enum {
    ePIMPR_SUBJECT                           ,
    ePIMPR_ALL_DAY_EVENT                     ,
    ePIMPR_SENSITIVITY                       ,
    ePIMPR_IS_RECURRING                      ,
    ePIMPR_REMINDER_SET                      ,
//  ePIMPR_REMINDER_SOUND_FILE               ,
    ePIMPR_REMINDER_OPTIONS                  ,
//  ePIMPR_DURATION                          ,
    ePIMPR_LOCATION                          ,
    ePIMPR_START                             ,
    ePIMPR_END                               ,
    ePIMPR_BUSY_STATUS                       ,
    ePIMPR_REMINDER_MINUTES_BEFORE_START     ,
//  ePIMPR_GLOBAL_OBJECT_ID                  ,
    ePIMPR_TIMEZONE                          ,
    ePIMPR_MEETING_STATUS                    ,
//  ePIMPR_MEETING_ORGANIZER_NAME            ,
//  ePIMPR_MEETING_OWNER_CRITICAL_CHANGE     ,
//  ePIMPR_ATTENDEES_CRITICAL_CHANGE         ,
    ePIMPR_CATEGORIES                        ,
    ePIMPR_BODY_TEXT                         ,
    
    // --- Recurring fields from here ---
    ePIMPR_RECURRING_TYPE                    ,
    ePIMPR_RECURRING_PATTERNSTARTDATE        ,
    ePIMPR_RECURRING_PATTERNENDDATE          ,
    ePIMPR_RECURRING_STARTTIME               ,
    ePIMPR_RECURRING_ENDTIME                 ,
    ePIMPR_RECURRING_NOEND                   ,
    ePIMPR_RECURRING_OCCURRENCES             ,
    ePIMPR_RECURRING_INTERVAL                ,
    ePIMPR_RECURRING_DAYOFWEEKMASK           ,
    ePIMPR_RECURRING_DAYOFMONTH              ,
    ePIMPR_RECURRING_INSTANCE                ,
//  ePIMPR_RECURRING_DURATION                ,
    ePIMPR_RECURRING_MONTHOFYEAR             ,

    // This is the total number of appointment properties supported (including recurring ones)
    NUM_COLS_APP
};


/**
 * Array of appointment fields used by WM plugin.
 * **********************************************
 *
 * Recurring properties are appended at the end.
 * @note Don't change the order of these properties!
 */
static CEPROPID PIMPR_props_app[] = {    
    PIMPR_SUBJECT                           ,
    PIMPR_ALL_DAY_EVENT                     ,
    PIMPR_SENSITIVITY                       ,
    PIMPR_IS_RECURRING                      ,
    PIMPR_REMINDER_SET                      ,
//  PIMPR_REMINDER_SOUND_FILE               ,
    PIMPR_REMINDER_OPTIONS                  ,
//  PIMPR_DURATION                          ,
    PIMPR_LOCATION                          ,
    PIMPR_START                             ,
    PIMPR_END                               ,
    PIMPR_BUSY_STATUS                       ,
    PIMPR_REMINDER_MINUTES_BEFORE_START     ,
//  PIMPR_GLOBAL_OBJECT_ID                  ,
    PIMPR_TIMEZONE                          ,
    PIMPR_MEETING_STATUS                    ,
//  PIMPR_MEETING_ORGANIZER_NAME            ,
//  PIMPR_MEETING_OWNER_CRITICAL_CHANGE     ,
//  PIMPR_ATTENDEES_CRITICAL_CHANGE         ,
    PIMPR_CATEGORIES                        ,
    PIMPR_BODY_TEXT                         ,
    
    // --- Recurring fields from here ---
    PIMPR_RECURRING_TYPE                    ,
    PIMPR_RECURRING_PATTERNSTARTDATE        ,
    PIMPR_RECURRING_PATTERNENDDATE          ,
    PIMPR_RECURRING_STARTTIME               ,
    PIMPR_RECURRING_ENDTIME                 ,
    PIMPR_RECURRING_NOEND                   ,
    PIMPR_RECURRING_OCCURRENCES             ,
    PIMPR_RECURRING_INTERVAL                ,
    PIMPR_RECURRING_DAYOFWEEKMASK           ,
    PIMPR_RECURRING_DAYOFMONTH              ,
    PIMPR_RECURRING_INSTANCE                ,
//  PIMPR_RECURRING_DURATION                ,
    PIMPR_RECURRING_MONTHOFYEAR
};


/**
 * Array of appointment fields names for SIF format.
 * Recurring fields are listed in recurrenceFields[] array.
 * @note Don't change the order of these properties!
 *       This list MUST be synchronized with PIMPR_props_app array of CEPROPID fields.
 */
const wchar_t* appointmentFields[] = {
    {L"Subject"                            },                // Returns or sets the subject for the Outlook item. This property corresponds to the MAPI property PR_SUBJECT. The Subject property is the default property for Outlook items.
    {L"AllDayEvent"                        },                // True if the appointment is an all-day event (as opposed to a specified time). Corresponds to the All day event check box on the Appointment page of an AppointmentItem - MUST be set AFTER "Start" and "End" fields.
    {L"Sensitivity"                        },                // Returns or sets the sensitivity for the Outlook item. Can be one of the following OlSensitivity constants: olConfidential(3), olNormal(0), olPersonal(1), or olPrivate(2). This property corresponds to the MAPI property PR_SENSITIVITY
    {L"IsRecurring"                        },                // True if the appointment or task is a recurring appointment or task. When the GetRecurrencePattern method is used with an AppointmentItem or TaskItem object, this property is set to True
    {L"ReminderSet"                        },                // True if a reminder has been set for this appointment, mail item or task. 
//  {L"ReminderSoundFile"                  },                // Returns or sets the path and filename of the sound file to play when the reminder occurs for the appointment or task. This property is only valid if the ReminderOverrideDefault and ReminderPlaySound properties are set to True
    {L"ReminderOptions"                    },                // What actions to take when the reminder occurs. (LED=1, Vibrate=2, Dialog=4, Sound=8, Repeat=16)
//  {L"Duration"                           },                // The length of an appointment in minutes.
    {L"Location"                           },                // Returns or sets the specific office location (for example, Building 1 Room 1 or Suite 123) for the appointment. This property corresponds to the MAPI property PR_OFFICE_LOCATION
    {L"Start"                              },                // Returns or sets the starting date and time for the appointment or journal entry. Use only in calendar
    {L"End"                                },                // Returns or sets the end date and time of an appointment or journal entry. Use only on calendar
    {L"BusyStatus"                         },                // Returns or sets the busy status of the user for the appointment. Can be one of the following OlBusyStatus constants: olBusy(2), olFree(0), olOutOfOffice(3), or olTentative(1). 
    {L"ReminderMinutesBeforeStart"         },                // Returns or sets the number of minutes the reminder should occur prior to the start of the appointment
    {L"Timezone"                           },                // The timezone of the appointment as a TIME_ZONE_INFORMATION structure
    {L"MeetingStatus"                      },                // OlRemoteStatus can be one of these OlRemoteStatus constants.
//  {L"MeetingOrganizerName"               },                // The name of the person who organized the meeting.
//  {L"MeetingOwnerCriticalChange"         },                // A timestamp indicating that the meeting owner wants to update the meeting request. Required when sending an updated meeting request to attendees.
//  {L"MeetingAttendeesCriticalChange"     },                // A timestamp indicating that a meeting attendee wants to update the meeting request. Required when sending an updated meeting request back to the meeting owner.
    {L"Categories"                         },                // Returns or sets the categories assigned to the Outlook item. 
    {L"Body"                               },                // Returns or sets the clear-text body of the Outlook item.
    {NULL}
};




/*
* Method to populate the appointment XML structure.
* It get every property name and value and build the XML structure to exchange with server.
*
* @param appointmentStringItem : the string that will contain the xml structure
* @param pAppointment          : the appointment object that contain every value
*
*/
void populateAppointmentStringItem(std::wstring &appointmentStringItem, IAppointment *pAppointment, BOOL isCRC) {

    BSTR element        = NULL      ;
    long elementLong    = 0         ;
    DATE startdate      = NULL      ;
    DATE enddate        = NULL      ;
    DATE date           = NULL      ;
    wstring element_s;

    //
    // FOR RECURRENCE
    // to get if a modification in the UTC goes into day before or after
    // store the initial value of day and check the last value
    //
    int x = 0, y = 0, z = 0;
    BOOL changeDay = FALSE, isBackward = FALSE;
    BOOL changeYearBackward = FALSE, changeYearForward = FALSE;


    VARIANT_BOOL variantBool = NULL; // 0==false -1==true
    wchar_t tempDate [DIM_LOCAL_TEMP];
    wchar_t* charBlob = NULL;

    appointmentStringItem = TEXT("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    appointmentStringItem +=  TEXT("<appointment>");

        // BSTR
        pAppointment->get_Subject(&element);
        element_s = wstring(element);
        encodeSpecialChar(element_s);
        appointmentStringItem  += TEXT("<Subject>") + element_s + TEXT("</Subject>");
        SysFreeString(element);

        pAppointment->get_Location(&element);
        element_s = wstring(element); encodeSpecialChar(element_s);
        appointmentStringItem  += TEXT("<Location>") + element_s + TEXT("</Location>");
        SysFreeString(element);

        pAppointment->get_Categories(&element);
        element_s = wstring(element); encodeSpecialChar(element_s);
        appointmentStringItem  +=  TEXT("<Categories>") + element_s + TEXT("</Categories>");
        SysFreeString(element);

        pAppointment->get_ReminderSoundFile(&element);
        element_s = wstring(element); encodeSpecialChar(element_s);
        appointmentStringItem  +=  TEXT("<ReminderSoundFile>") + element_s + TEXT("</ReminderSoundFile>");
        SysFreeString(element);

        pAppointment->get_Body(&element);
        removeEndCarriage(&element);
        element_s = wstring(element);  encodeSpecialChar(element_s);
        appointmentStringItem  +=  TEXT("<Body>") + element_s + TEXT("</Body>");
        SysFreeString(element);

        // DATE
        pAppointment->get_Start(&startdate);
        pAppointment->get_End(&enddate);


        /*
        * see the start and end date. If the start is 00:00 and end is 23:59 the appointment is decided to be
        * an all day event. So the AllDayEvent tag is put to 'true' only in the SIF structure. The format of the date
        * is like the birthday: yyyy-MM-dd
        *
        */

        BOOL isAllDay = FALSE;

        if (isWindowsMobile5()) {
            pAppointment->get_AllDayEvent(&variantBool);
            variantBool == -1 ? isAllDay = TRUE : isAllDay = FALSE;

            if (isAllDay)
                enddate -= 1; // to be decremented of a day if it is all day in windows mobile 5.
                // This because the end date is 00:00:00 of the next day. we want the all day
                // is represented as <startDate>2007-01-15</StartDate> and and date as the same

        } else {
            isAllDayEvent(startdate, enddate);
        }


        wsprintf(tempDate, TEXT(""));
        double toCompare = MAX_DATE_DOUBLE;
        double toCompare4000 = MAX_DATE_DOUBLE_4000;

        SYSTEMTIME st;

        if (startdate != toCompare) {
            VariantTimeToSystemTime(startdate, &st);
            x = st.wHour;
            localTimeToUTC2(st);
            y = st.wHour;
            if (isAllDay) {
                doubleToSystemTimeBirthday(tempDate, startdate);
            } else {
                doubleToSystemTime(tempDate, startdate);
            }
        }
        appointmentStringItem  = appointmentStringItem + TEXT("<Start>") + tempDate + TEXT("</Start>");

        wsprintf(tempDate, TEXT(""));
        if (enddate != toCompare) {
            if (isAllDay) {
                doubleToSystemTimeBirthday(tempDate, enddate);
            } else {
                doubleToSystemTime(tempDate, enddate);
            }
        }
        appointmentStringItem  = appointmentStringItem + TEXT("<End>") + tempDate + TEXT("</End>");

        if (isAllDay) {
            wsprintf(tempDate, TEXT("1")); // TRUE
        } else {
            wsprintf(tempDate, TEXT("0")); // TRUE
        }
        appointmentStringItem  = appointmentStringItem + TEXT("<AllDayEvent>") + tempDate + TEXT("</AllDayEvent>");

        pAppointment->get_ReminderSet(&variantBool);
        wsprintf(tempDate, TEXT("%i"), variantBool);
        normalizeBoolean(tempDate);
        appointmentStringItem  = appointmentStringItem + TEXT("<ReminderSet>") + tempDate + TEXT("</ReminderSet>");

        // long
        //pAppointment->get_Duration(&elementLong);
        //wsprintf(tempDate, TEXT("%i"), elementLong);
        //appointmentStringItem  = appointmentStringItem + TEXT("<Duration>") + tempDate + TEXT("</Duration>");

        pAppointment->get_BusyStatus(&elementLong);
        wsprintf(tempDate, TEXT("%i"), elementLong);
        appointmentStringItem  = appointmentStringItem + TEXT("<BusyStatus>") + tempDate + TEXT("</BusyStatus>");

        pAppointment->get_Sensitivity(&elementLong);
        wsprintf(tempDate, TEXT("%i"), elementLong);
        appointmentStringItem  = appointmentStringItem + TEXT("<Sensitivity>") + tempDate + TEXT("</Sensitivity>");

        pAppointment->get_ReminderOptions(&elementLong);
        wsprintf(tempDate, TEXT("%i"), elementLong);
        appointmentStringItem  = appointmentStringItem + TEXT("<ReminderOptions>") + tempDate + TEXT("</ReminderOptions>");

        pAppointment->get_ReminderMinutesBeforeStart(&elementLong);
        wsprintf(tempDate, TEXT("%i"), elementLong);
        appointmentStringItem  = appointmentStringItem + TEXT("<ReminderMinutesBeforeStart>") + tempDate + TEXT("</ReminderMinutesBeforeStart>");


        /******************************************************/

        //Recurrence
        VARIANT_BOOL isRecurring;
        pAppointment->get_IsRecurring(&isRecurring);
        wsprintf(tempDate, TEXT("%i"), isRecurring);
        normalizeBoolean(tempDate);
        appointmentStringItem  = appointmentStringItem + TEXT("<IsRecurring>") + tempDate + TEXT("</IsRecurring>");

        IRecurrencePattern *pRecurrence;
        pAppointment->GetRecurrencePattern (&pRecurrence);

        wstring recurrence = L"";
        if (!isCRC || isRecurring == -1) { // variant bool to say TRUE
            recurrence = getRecurrenceTags(pRecurrence, isRecurring, x, y, olFolderCalendar, startdate, isAllDay);
        }
        // handle recurrence and exceptions events
        // wstring recurrence = getRecurrenceTags(pRecurrence, isRecurring, x, y, olFolderCalendar, startdate, isAllDay);

        pRecurrence->Release();

    appointmentStringItem += recurrence;
    appointmentStringItem = appointmentStringItem +  TEXT("</appointment>");

}




BOOL isAllDayEvent(DATE startdate, DATE enddate) {

    SYSTEMTIME ststart, stend;
    VariantTimeToSystemTime(startdate, &ststart);
    VariantTimeToSystemTime(enddate, &stend);
    BOOL ret = FALSE;

    if (ststart.wHour == 0 && ststart.wMinute == 0 &&
        stend.wHour == 23 && stend.wMinute == 59)

        ret = TRUE;

    return ret;

}




void copyFields(IAppointment *pAppointmentExc, IAppointment *pAppointmentCopy) {

    BSTR element = NULL;

    pAppointmentExc->get_Subject(&element);
    pAppointmentCopy->put_Subject(element);
    SysFreeString(element);

    pAppointmentExc->get_Location(&element);
    pAppointmentCopy->put_Location(element);
    SysFreeString(element);

    pAppointmentExc->get_Body(&element);
    pAppointmentCopy->put_Body(element);
    SysFreeString(element);

    DATE d;
    pAppointmentExc->get_Start(&d);
    pAppointmentCopy->put_Start(d);

    pAppointmentExc->get_End(&d);
    pAppointmentCopy->put_End(d);

    VARIANT_BOOL vbool;
    pAppointmentExc->get_AllDayEvent(&vbool);
    pAppointmentCopy->put_AllDayEvent(vbool);

    pAppointmentExc->get_ReminderSet(&vbool);
    pAppointmentCopy->put_ReminderSet(vbool);

    long elementLong = 0;
    pAppointmentExc->get_BusyStatus(&elementLong);
    pAppointmentCopy->put_BusyStatus(elementLong);

    pAppointmentExc->get_ReminderMinutesBeforeStart(&elementLong);
    pAppointmentCopy->put_ReminderMinutesBeforeStart(elementLong);

}


/*
* . Scan all the appointment and get the ones with recurrences
* . If there are any that are with exceptions remove them
* .
*/

BOOL normalizeAppointmentExceptions(IP_OUTLOOK_APP* polApp) {

    IPOutlookItemCollection *pItems;
    IAppointment *pAppointment;
    IFolder *pFolder;

    VARIANT_BOOL variantBool = FALSE;

    polApp->GetDefaultFolder (olFolderCalendar, &pFolder);
    pFolder->get_Items (&pItems);

    int currentCountItems = 0;

    pItems->get_Count(&currentCountItems);

    for (int i = 0; i < currentCountItems; i++) {
        VARIANT_BOOL isRecurring;
        pItems->Item (i + 1, (IDispatch**)&pAppointment);
        pAppointment->get_IsRecurring(&isRecurring);

        if (isRecurring >= 0) { // VARIANT_BOOL true is -1. If not recurring continue
            pAppointment->Release();
            continue;
        }

        IRecurrencePattern *pRecurrence;
        pAppointment->GetRecurrencePattern (&pRecurrence);

        IExceptions* pExceptions;
        int numExceptions;

        pRecurrence->get_Exceptions(&pExceptions);

        if (!pExceptions) { // there are no exception in this occurrence
            continue;
        }

        pExceptions->get_Count(&numExceptions);

        for (int i = 0; i < numExceptions; i++) {
            IException* pException;
            pExceptions->Item(i + 1, &pException);

            // if the recurrence is deleted get the deletion and continue to the next appointment
            VARIANT_BOOL isDeleted;
            pException->get_Deleted(&isDeleted);
            if (isDeleted == -1) {
                pException->Release();
                continue;
            }

            IAppointment* pAppointmentExc, *pAppointmentCopy, *pAppointmentNew;
            pException->get_AppointmentItem(&pAppointmentExc);

            HRESULT h = pAppointment->Copy(&pAppointmentCopy);

            // copy the fields from the pAppintmentExc to the pAppointmentCopy
            copyFields(pAppointmentExc, pAppointmentCopy);

            pAppointmentExc->Delete();
            pAppointmentExc->Release();

            pAppointmentCopy->Save();

            /*
            * There is an exception in ClearRecurrencePattern on the copy appointment.
            * it is safe to open a new appointment object
            */
            long oid = 0;
            h = pAppointmentCopy->get_Oid(&oid);
            pAppointmentCopy->Release();

            polApp->GetItemFromOid (oid, (IDispatch**)&pAppointmentNew);
            if (pAppointmentNew) {
                h = pAppointmentNew->ClearRecurrencePattern();
                pAppointmentNew->Save();
                pAppointmentNew->Release();
            } else {
                LOG.error("Appointment normalization: it is impossible to clear the pattern %i", oid);
            }
            pException->Release();

        }
        pExceptions->Release();
        pRecurrence->Release();

        pAppointment->Release();
    }

    pFolder->Release();
    pItems->Release();

    return TRUE;
}

/*
* The method calculate the CRC hash value for the appointment item.
* It create the XML structure and calculate the crc value.
* Use to compare for understand the modified value!
*
* @param pAppointment : the appointment object that contain every value
*/

long calculateAppointmentHash (IAppointment* pAppointment) {

    wstring p;
    populateAppointmentStringItem(p, pAppointment, TRUE);

    long l = stringCrc32(p.data());

    return l;
}





WinItem* appointmentToWinEvent(IItem *pAppointment, bool isSif) {

    //MessageBox (NULL, TEXT("appbuilder"), TEXT("title"), MB_SETFOREGROUND |MB_OK);

    HRESULT hr                      = E_FAIL;
    CEPROPVAL *prgPropvalPoom       = NULL;
    ULONG cbBuffer                  = 0;
    HANDLE hHeap                    = GetProcessHeap();
    WinEvent* winE                  = NULL;
    WinRecurrence* winRec           = NULL;
    IAppointment* pAppointmentOld   = NULL;
    IRecurrencePattern* pRecurrence = NULL;
    IExceptions* pExceptions        = NULL;
    int numExceptions               = 0;
    bool isAllDay                   = false;
    bool isRecurring                = false;
    wchar_t tempValue[10];
    wstring value, start;

    if (isSif) { winE = new WinEventSIF(); }
    else       { winE = new WinEvent()   ; }

    cbBuffer = 0;
    hr = pAppointment->GetProps(PIMPR_props_app, CEDB_ALLOWREALLOC, NUM_COLS_APP, &prgPropvalPoom, &cbBuffer, hHeap);
    if(FAILED(hr) || NULL == prgPropvalPoom || 0 == cbBuffer) {
        goto Exit;
    }


    //
    // Fill all appointment properties
    //
    for (int i = 0; i < ePIMPR_RECURRING_TYPE; i++) {

        switch (i)
        {
        case ePIMPR_ALL_DAY_EVENT:
            {
                int tmp = getIntPropertyFromMapi(&prgPropvalPoom[i]);
                wsprintf(tempValue, TEXT("%d"), tmp);
                value = tempValue;
                if (value == TEXT("1")) {
                    isAllDay = true;
                }
                break;
            }
        case ePIMPR_IS_RECURRING:
            {
                int tmp = getIntPropertyFromMapi(&prgPropvalPoom[i]);
                wsprintf(tempValue, TEXT("%d"), tmp);
                value = tempValue;
                if (value == TEXT("1")) {
                    isRecurring = true;
                }
                break;
            }
        case ePIMPR_START:
        case ePIMPR_END:
            {
                // Date values: "yyyyMMdd" or "YYYYMMDDThhmmssZ"
                value = getDatePropertyFromMapi(&prgPropvalPoom[i], !isAllDay, isAllDay);
                break;
            }
        case ePIMPR_SENSITIVITY:
        case ePIMPR_REMINDER_SET:
        case ePIMPR_REMINDER_OPTIONS:
      //case ePIMPR_DURATION:
        case ePIMPR_BUSY_STATUS:
        case ePIMPR_REMINDER_MINUTES_BEFORE_START:
        case ePIMPR_MEETING_STATUS:
            {
                // Int values.
                int tmp = getIntPropertyFromMapi(&prgPropvalPoom[i]);
                wsprintf(tempValue, TEXT("%d"), tmp);
                value = tempValue;
                break;
            }
        case ePIMPR_TIMEZONE:
            {
                // Timezone: if set, recurring props are sent in local time.
                // We pass the tz structure to API.
                TIME_ZONE_INFORMATION* tz = readTimezoneFromMapi(&prgPropvalPoom[i]);
                if (tz) {
                    winE->setTimezone(tz);
                    delete [] tz;
                }
                continue;
            }
        default:
            {
                // String values.
                value = getStringPropertyFromMapi(&prgPropvalPoom[i]);
                break;
            }
        }

        winE->setProperty(appointmentFields[i], value);
        //LOG.info("%2d) %ls = \"%ls\"", i+1, appointmentFields[i], value.c_str());
    }


    if (!isRecurring) {
        goto Exit;
    }


    //
    // Manage recurring properties: need to retrieve all values together
    // ---------------------------- to apply safe checks and conversions.
    //
    winRec = winE->getRecPattern();
    if (!winRec) {
        goto Exit;
    }
    winE->getProperty(TEXT("Start"), start);
    setWinRecurrenceProps(*winRec, prgPropvalPoom, ePIMPR_RECURRING_TYPE, start, isAllDay, true);



    //
    // Exceptions. Use old style IAppointment interface.
    // -----------
    long oid = 0;
    hr = pAppointment->get_Oid(&oid);
    if (!oid || FAILED(hr))             { goto Exit; }

    IP_OUTLOOK_APP* pOutlookApp = getOutlookApp();
    if (!pOutlookApp)                   { goto Exit; }

    hr = pOutlookApp->GetItemFromOid(oid, (IDispatch**)&pAppointmentOld);
    if (!pAppointmentOld || FAILED(hr)) { goto Exit; }

    hr = pAppointmentOld->GetRecurrencePattern(&pRecurrence);
    if (!pRecurrence || FAILED(hr))    { goto Exit; }

    hr = pRecurrence->get_Exceptions(&pExceptions);
    if (!pExceptions || FAILED(hr))    { goto Exit; }

    pExceptions->get_Count(&numExceptions);
    if (numExceptions == 0) {
        goto Exit;
    }

    DATE startDate = NULL;
    systemTimeToDouble(start.c_str(), &startDate, NULL);

    // Push all EXDATEs into WinEvent list (we have only EXDATEs)
    for (int i=0; i < numExceptions; i++) {
        IException* pException;
        hr = pExceptions->Item(i + 1, &pException);
        if (!pException || FAILED(hr)) { goto Exit; }

        wstring exception = createException(pException, startDate, isAllDay);
        winE->getExcludeDates()->push_back(exception);
        pException->Release();
    }


Exit:
    // Release objects.
    if (pExceptions)     { pExceptions->Release();    }
    if (pRecurrence)     { pRecurrence->Release();    }
    if (pAppointmentOld) { pAppointmentOld->Release();}

    // Free memory.
    HeapFree(hHeap, 0, prgPropvalPoom);
    return winE;
}




HRESULT winEventToAppointment(WinEvent& winE, IItem *pAppointment, const wstring& path, long oidToUpdate) {

    HRESULT hr                      = E_FAIL;
    int incNumProp                  = 0;
    bool exists                     = false;
    WinRecurrence* winRec           = NULL;
    IAppointment* pAppointmentOld   = NULL;
    IRecurrencePattern* pRecurrence = NULL;
    CEPROPVAL *prgPropvalPoom       = new CEPROPVAL[NUM_COLS_APP];      // Warning: allocation of CEPROPVAL array is critical...
    IP_OUTLOOK_APP* pOutlookApp     = NULL;
    FILETIME date;
    wstring start, end;

    vector<RecurrenceException> listExceptions;
    list<wstring>::iterator it;
    vector<RecurrenceException>::iterator exc;
    vector<long> oids;
    vector<long>::iterator oidsIterator;
    int numExceptions               = 0;
    IExceptions* pExceptions        = NULL;


    if (oidToUpdate) {
        //
        // It's an UPDATED item: Clear the recurrence pattern & save event: 
        // if the event is recurring or not, we have to delete any previous information 
        // regarding recurrence and exceptions.
        //
        pOutlookApp = getOutlookApp();
        if (!pOutlookApp) { goto finally; }

        // 'pAppointmentOld' is the IAppointment (old interface)
        hr = pOutlookApp->GetItemFromOid(oidToUpdate, (IDispatch**)&pAppointmentOld);
        if (!pAppointmentOld || FAILED(hr)) { goto finally; }

        hr = pAppointmentOld->ClearRecurrencePattern();
        if (FAILED(hr)) {
            LOG.error("Error clearing rec pattern");
        }

        hr = pAppointmentOld->Save();
        if (FAILED(hr)) {
            LOG.error("Error saving appointment \"%ls\"", winE.getName().c_str());
            goto finally;
        }
        pAppointmentOld->Release();
        pAppointmentOld = NULL;
    }



    //
    // Check if it's an all-day event / recurring.
    //
    bool isAllDay = false, isRecurring = false;
    int tmp = 0;
    if (winE.getIntProperty(TEXT("AllDayEvent"), &tmp)) {
        if (tmp == 1) isAllDay = true;
    }
    if (winE.getIntProperty(TEXT("IsRecurring"), &tmp)) {
        if (tmp == 1) isRecurring = true;
    }


    //
    // Set all normal properties.
    //
    for (int i=0; i < ePIMPR_RECURRING_TYPE; i++) {
        exists = false;
        wstring& value = winE.getPropertyRef(appointmentFields[i], &exists);

        if (i == ePIMPR_TIMEZONE) {
            // Correction: timezone is stored into a tzInfo structure.
            exists = winE.hasTimezone();
        }

        if (exists) {
            switch (i)
            {
            case ePIMPR_START:
            case ePIMPR_END:
                {
                    // Date values: "yyyyMMdd" or "YYYYMMDDThhmmssZ"
                    date = getDatePropertyFromString(value, isAllDay);
                    setDatePropertyToMapi(&prgPropvalPoom[incNumProp], PIMPR_props_app[i], date);
                    //LOG.debug("%2d. %ls = %ls (date) (%lu %lu)", incNumProp+1, appointmentFields[i], value.c_str(), date.dwHighDateTime, date.dwLowDateTime);
                    break;
                }
            case ePIMPR_ALL_DAY_EVENT:
            case ePIMPR_SENSITIVITY:
            case ePIMPR_IS_RECURRING:
            case ePIMPR_REMINDER_SET:
            case ePIMPR_REMINDER_OPTIONS:
          //case ePIMPR_DURATION:
            case ePIMPR_BUSY_STATUS:
            case ePIMPR_REMINDER_MINUTES_BEFORE_START:
            case ePIMPR_MEETING_STATUS:
                {
                    // Int values: should add safe checks?                  ***** TODO ******
                    int val = getIntPropertyFromWinItem(value);
                    setIntPropertyToMapi(&prgPropvalPoom[incNumProp], PIMPR_props_app[i], val);
                    //LOG.debug("%2d. %ls = %d (int)", incNumProp+1, appointmentFields[i], val);
                    break;
                }
            case ePIMPR_TIMEZONE:
                {
                    // Timezone
                    setTimezoneToMapi(&prgPropvalPoom[incNumProp], PIMPR_props_app[i], winE.getTimezone());
                    break;
                }
            case ePIMPR_BODY_TEXT:
                {
                    convertNewlines(value);     // BUG #1562
                    setStringPropertyToMapi(&prgPropvalPoom[incNumProp], PIMPR_props_app[i], value);
                    //LOG.debug("%2d. %ls = %ls (body)", incNumProp+1, appointmentFields[i], value.c_str());
                    break;
                }
            default:
                {
                    // String values.
                    setStringPropertyToMapi(&prgPropvalPoom[incNumProp], PIMPR_props_app[i], value);
                    //LOG.debug("%2d. %ls = %ls (string)", incNumProp+1, appointmentFields[i], value.c_str());
                    break;
                }
            }
            incNumProp++;
        }
    }


    if (isRecurring) {
        //
        // Manage recurring properties: need to retrieve all values together
        // ---------------------------- to apply safe checks and conversions.
        //
        winRec = winE.getRecPattern();
        if (!winRec) { goto finally; }

        winE.getProperty(TEXT("Start"), start);
        winE.getProperty(TEXT("End"),   end);
        incNumProp = setMAPIRecurrenceProps(prgPropvalPoom, incNumProp, *winRec, start, 
                                            end, isAllDay, winE.getTimezone(), true);
    }


    //
    // Save all props together (with rec props if recurring).
    //
    hr = saveItemProps(pAppointment, incNumProp, prgPropvalPoom);
    if (FAILED(hr)) {
        LOG.error("Error saving appointment \"%ls\"", winE.getName().c_str());
        goto finally;
    }

    if (!isRecurring) {
        // Not recurring: finished.
        goto finally;
    }


    //
    // Get again the IAppointment to manage exceptions.
    //
    long appOid = 0;
    hr = pAppointment->get_Oid(&appOid);
    if (!appOid || FAILED(hr))          { goto finally; }

    if (!pOutlookApp) {
        pOutlookApp = getOutlookApp();
        if (!pOutlookApp)               { goto finally; }
    }
    hr = pOutlookApp->GetItemFromOid(appOid, (IDispatch**)&pAppointmentOld);
    if (!pAppointmentOld || FAILED(hr)) { goto finally; }

    hr = pAppointmentOld->GetRecurrencePattern(&pRecurrence);
    if (!pRecurrence || FAILED(hr))     { goto finally; }


    // Safe check?
    //if (pRecurrence->get_Exceptions(&pExceptions)) {
    //    if (pExceptions->get_Count(&numExceptions)) {
    //        if (numExceptions > 0) {
    //            LOG.info("Warning: event still have %d exceptions...", numExceptions);
    //        }
    //    }
    //}


    //
    // Manage appointment exceptions (old style).
    // ------------------------------------------
    // note: MUST set exceptions AFTER the item is saved.
    // note: MUST clear rec pattern before handling exceptions.
    //
    exceptionList* exDates = winE.getExcludeDates();
    exceptionList* inDates = winE.getIncludeDates();
    if (!exDates->size() && !inDates->size()) {
        // No exceptions.
        goto finally;
    }


    // Fill listExceptions with EXDATEs
    listExceptions.clear();
    for (it = exDates->begin(); it != exDates->end(); it++) {
        RecurrenceException rec;
        DATE originalDate = 0;
        LOG.debug("originalDate = %ls (allday = %d)", (*it).c_str(), isAllDay);

        if (isAllDay) { systemTimeToDoubleBirthFormat((*it).c_str(), &originalDate, TEXT("START")); }
        else          { systemTimeToDouble           ((*it).c_str(), &originalDate, NULL         ); }

        rec.setOriginalDate(originalDate);
        rec.setIsDeleted(TRUE);
        listExceptions.push_back(rec);
    }

    // Fill listExceptions with RDATEs
    for (it = inDates->begin(); it != inDates->end(); it++) {
        RecurrenceException rec;
        DATE originalDate;
        LOG.debug("originalDate = %ls (allday = %d)", (*it).c_str(), isAllDay);

        if (isAllDay) { systemTimeToDoubleBirthFormat((*it).c_str(), &originalDate, TEXT("START")); }
        else          { systemTimeToDouble           ((*it).c_str(), &originalDate, NULL         ); }

        rec.setOriginalDate(originalDate);
        rec.setIsDeleted(FALSE);
        rec.setAllDayEvent(isAllDay);
        listExceptions.push_back(rec);
    }


    LOG.debug("Adding exceptions...");
    for (exc = listExceptions.begin(); exc != listExceptions.end(); exc++) {
        if ((*exc).isDeleted()) {
            //
            // EXDATE: delete the occurrence
            //
            IAppointment* pExAppointment;
            hr = pRecurrence->GetOccurrence((*exc).getOriginalDate(), &pExAppointment);
            if(!pExAppointment || FAILED(hr)) {
                LOG.error("Error in handling calendar exception. pExAppointment is NULL. Date: %f", (*exc).getOriginalDate());
                continue;
            }
            pExAppointment->Delete();
            pExAppointment->Release();
        }

        else {
            //
            // RDATE: not supported by pOutlook. Create a new indipendent event and
            // add recurring event to 'oidToReset' list (will be sent as modified next time)
            //
            oids.push_back(appOid);
            long oid = createAppointmentFromException(pAppointmentOld, *exc, pOutlookApp);
            oids.push_back(oid);
        }
    }

    if (oids.size() > 0) {
        writeOidToReset(oids, path.c_str());
    }

    hr = S_OK;


finally:

    // Release objects.
    if (pRecurrence)     { pRecurrence->Release();     }
    if (pAppointmentOld) { pAppointmentOld->Release(); }

    if (prgPropvalPoom) {
        delete [] prgPropvalPoom;
    }

    //LOG.debug("Exiting winEventToAppointment");
    return hr;
}

ArrayList* getAppointmentProperties(){

    int i = 0;
    ArrayList* props = new ArrayList();

    while (appointmentFields[i]){
        Property prop;
        char* field = toMultibyte(appointmentFields[i]);
        prop.setPropName(field);
        props->add(prop);
        
        delete [] field;

        i++;
    }

    return props;
}

ArrayList* getVCalendarProperties(){

    ArrayList* props = new ArrayList();
    Property* p = new Property();
    ArrayList valenums;
    StringBuffer val;


    p->setPropName("BEGIN");
    val = "VCALENDAR";
    valenums.add(val);
    p->setValEnums(&valenums);
    props->add(*p);

    valenums.clear();

    p->setPropName("END");
    val = "VCALENDAR";
    valenums.add(val);
    p->setValEnums(&valenums);
    props->add(*p);

    valenums.clear();
    
    p->setPropName("VERSION");
    val = "1.0";
    valenums.add(val);
    p->setValEnums(&valenums);
    props->add(*p);

    p->setPropName("X-FUNAMBOL-ALLDAY");
    props->add(*p);

    p->setPropName("DESCRIPTION");
    props->add(*p);
    p->setPropName("X-MICROSOFT-CDO-BUSYSTATUS");
    props->add(*p);
    p->setPropName("CATEGORIES");
    props->add(*p);
    p->setPropName("DTEND");
    props->add(*p);
    p->setPropName("LOCATION");
    props->add(*p);
    p->setPropName("STATUS");
    props->add(*p);
    p->setPropName("X-FUNAMBOL-AALARMOPTIONS");
    props->add(*p);
    p->setPropName("AALARM");
    props->add(*p);
    p->setPropName("CLASS");
    //TODO: Values?
    props->add(*p);
    p->setPropName("DTSTART");
    props->add(*p);
    p->setPropName("SUMMARY");
    props->add(*p);
    p->setPropName("RRULE");
    props->add(*p);
    p->setPropName("EXDATE");
    props->add(*p);
    p->setPropName("RDATE");
    props->add(*p);
    p->setPropName("TZ");
    props->add(*p);
    p->setPropName("DAYLIGHT");
    props->add(*p);

    delete p; p = NULL;

    return props;
}